# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This code is refered from:
https://github.com/WenmuZhou/DBNet.pytorch/blob/master/post_processing/seg_detector_representer.py
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import cv2
# import paddle
from shapely.geometry import Polygon
import pyclipper


class DBPostProcess(object):
  """
  The post process for Differentiable Binarization (DB).
  """

  def __init__(self,
         thresh=0.3,
         box_thresh=0.7,
         max_candidates=1000,
         unclip_ratio=2.0,
         use_dilation=False,
         score_mode="fast",
         box_type='quad',
         **kwargs):
    self.thresh = thresh
    self.box_thresh = box_thresh
    self.max_candidates = max_candidates
    self.unclip_ratio = unclip_ratio
    self.min_size = 3
    self.score_mode = score_mode
    self.box_type = box_type
    assert score_mode in [
      "slow", "fast"
    ], "Score mode must be in [slow, fast] but got: {}".format(score_mode)

    self.dilation_kernel = None if not use_dilation else np.array(
      [[1, 1], [1, 1]])

  def polygons_from_bitmap(self, pred, _bitmap, dest_width, dest_height):
    '''
    _bitmap: single map with shape (1, H, W),
      whose values are binarized as {0, 1}
    '''

    bitmap = _bitmap
    height, width = bitmap.shape

    boxes = []
    scores = []

    contours, _ = cv2.findContours((bitmap * 255).astype(np.uint8),
                     cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)

    for contour in contours[:self.max_candidates]:
      epsilon = 0.002 * cv2.arcLength(contour, True)
      approx = cv2.approxPolyDP(contour, epsilon, True)
      points = approx.reshape((-1, 2))
      if points.shape[0] < 4:
        continue

      score = self.box_score_fast(pred, points.reshape(-1, 2))
      if self.box_thresh > score:
        continue

      if points.shape[0] > 2:
        box = self.unclip(points, self.unclip_ratio)
        if len(box) > 1:
          continue
      else:
        continue
      box = box.reshape(-1, 2)

      _, sside = self.get_mini_boxes(box.reshape((-1, 1, 2)))
      if sside < self.min_size + 2:
        continue

      box = np.array(box)
      box[:, 0] = np.clip(
        np.round(box[:, 0] / width * dest_width), 0, dest_width)
      box[:, 1] = np.clip(
        np.round(box[:, 1] / height * dest_height), 0, dest_height)
      boxes.append(box.tolist())
      scores.append(score)
    return boxes, scores

  def boxes_from_bitmap(self, pred, _bitmap, dest_width, dest_height):
    '''
    _bitmap: single map with shape (1, H, W),
        whose values are binarized as {0, 1}
    '''

    bitmap = _bitmap
    height, width = bitmap.shape

    outs = cv2.findContours((bitmap * 255).astype(np.uint8), cv2.RETR_LIST,
                cv2.CHAIN_APPROX_SIMPLE)
    if len(outs) == 3:
      img, contours, _ = outs[0], outs[1], outs[2]
    elif len(outs) == 2:
      contours, _ = outs[0], outs[1]

    num_contours = min(len(contours), self.max_candidates)

    boxes = []
    scores = []
    for index in range(num_contours):
      contour = contours[index]
      points, sside = self.get_mini_boxes(contour)
      if sside < self.min_size:
        continue
      points = np.array(points)
      if self.score_mode == "fast":
        score = self.box_score_fast(pred, points.reshape(-1, 2))
      else:
        score = self.box_score_slow(pred, contour)
      if self.box_thresh > score:
        continue

      box = self.unclip(points, self.unclip_ratio).reshape(-1, 1, 2)
      box, sside = self.get_mini_boxes(box)
      if sside < self.min_size + 2:
        continue
      box = np.array(box)

      box[:, 0] = np.clip(
        np.round(box[:, 0] / width * dest_width), 0, dest_width)
      box[:, 1] = np.clip(
        np.round(box[:, 1] / height * dest_height), 0, dest_height)
      boxes.append(box.astype("int32"))
      scores.append(score)
    return np.array(boxes, dtype="int32"), scores

  def unclip(self, box, unclip_ratio):
    poly = Polygon(box)
    distance = poly.area * unclip_ratio / poly.length
    offset = pyclipper.PyclipperOffset()
    offset.AddPath(box, pyclipper.JT_ROUND, pyclipper.ET_CLOSEDPOLYGON)
    expanded = np.array(offset.Execute(distance))
    return expanded

  def get_mini_boxes(self, contour):
    bounding_box = cv2.minAreaRect(contour)
    points = sorted(list(cv2.boxPoints(bounding_box)), key=lambda x: x[0])

    index_1, index_2, index_3, index_4 = 0, 1, 2, 3
    if points[1][1] > points[0][1]:
      index_1 = 0
      index_4 = 1
    else:
      index_1 = 1
      index_4 = 0
    if points[3][1] > points[2][1]:
      index_2 = 2
      index_3 = 3
    else:
      index_2 = 3
      index_3 = 2

    box = [
      points[index_1], points[index_2], points[index_3], points[index_4]
    ]
    return box, min(bounding_box[1])

  def box_score_fast(self, bitmap, _box):
    '''
    box_score_fast: use bbox mean score as the mean score
    '''
    h, w = bitmap.shape[:2]
    box = _box.copy()
    xmin = np.clip(np.floor(box[:, 0].min()).astype("int32"), 0, w - 1)
    xmax = np.clip(np.ceil(box[:, 0].max()).astype("int32"), 0, w - 1)
    ymin = np.clip(np.floor(box[:, 1].min()).astype("int32"), 0, h - 1)
    ymax = np.clip(np.ceil(box[:, 1].max()).astype("int32"), 0, h - 1)

    mask = np.zeros((ymax - ymin + 1, xmax - xmin + 1), dtype=np.uint8)
    box[:, 0] = box[:, 0] - xmin
    box[:, 1] = box[:, 1] - ymin
    cv2.fillPoly(mask, box.reshape(1, -1, 2).astype("int32"), 1)
    return cv2.mean(bitmap[ymin:ymax + 1, xmin:xmax + 1], mask)[0]

  def box_score_slow(self, bitmap, contour):
    '''
    box_score_slow: use polyon mean score as the mean score
    '''
    h, w = bitmap.shape[:2]
    contour = contour.copy()
    contour = np.reshape(contour, (-1, 2))

    xmin = np.clip(np.min(contour[:, 0]), 0, w - 1)
    xmax = np.clip(np.max(contour[:, 0]), 0, w - 1)
    ymin = np.clip(np.min(contour[:, 1]), 0, h - 1)
    ymax = np.clip(np.max(contour[:, 1]), 0, h - 1)

    mask = np.zeros((ymax - ymin + 1, xmax - xmin + 1), dtype=np.uint8)

    contour[:, 0] = contour[:, 0] - xmin
    contour[:, 1] = contour[:, 1] - ymin

    cv2.fillPoly(mask, contour.reshape(1, -1, 2).astype("int32"), 1)
    return cv2.mean(bitmap[ymin:ymax + 1, xmin:xmax + 1], mask)[0]

  def __call__(self, outs_dict, shape_list):
    pred = outs_dict['maps']
    # if isinstance(pred, paddle.Tensor):
    #   pred = pred.numpy()
    pred = pred[:, 0, :, :]
    segmentation = pred > self.thresh

    boxes_batch = []
    for batch_index in range(pred.shape[0]):
      src_h, src_w, ratio_h, ratio_w = shape_list[batch_index]
      if self.dilation_kernel is not None:
        mask = cv2.dilate(
          np.array(segmentation[batch_index]).astype(np.uint8),
          self.dilation_kernel)
      else:
        mask = segmentation[batch_index]
      if self.box_type == 'poly':
        boxes, scores = self.polygons_from_bitmap(pred[batch_index],
                              mask, src_w, src_h)
      elif self.box_type == 'quad':
        boxes, scores = self.boxes_from_bitmap(pred[batch_index], mask,
                             src_w, src_h)
      else:
        raise ValueError("box_type can only be one of ['quad', 'poly']")

      boxes_batch.append({'points': boxes})
    return boxes_batch



import numpy as np

from typing import Iterable
from shapely.geometry import Polygon
from PIL.Image import new, Image, Resampling
from .types import Layout, OCRFragment
from .ocr import OCR
from .overlap import overlap_rate
from .rectangle import Point, Rectangle


_MIN_RATE = 0.5

def correct_fragments(ocr: OCR, source: Image, layout: Layout):
  x1, y1, x2, y2 = layout.rect.wrapper
  image: Image = source.crop((
    round(x1), round(y1),
    round(x2), round(y2),
  ))
  image, dx, dy, scale = _adjust_image(image)
  image_np = np.array(image)
  ocr_fragments = list(ocr.search_fragments(image_np))
  corrected_fragments: list[OCRFragment] = []

  for fragment in ocr_fragments:
    _apply_fragment(fragment.rect, layout, dx, dy, scale)

  matched_fragments, not_matched_fragments = _match_fragments(
    zone_rect=layout.rect,
    fragments1=layout.fragments,
    fragments2=ocr_fragments,
  )
  for fragment1, fragment2 in matched_fragments:
    if fragment1.rank > fragment2.rank:
      corrected_fragments.append(fragment1)
    else:
      corrected_fragments.append(fragment2)

  corrected_fragments.extend(not_matched_fragments)
  layout.fragments = corrected_fragments

def _adjust_image(image: Image) -> tuple[Image, int, int, float]:
  # after testing, adding white borders to images can reduce
  # the possibility of some text not being recognized
  border_size: int = 50
  adjusted_size: int = 1024 - 2 * border_size
  width, height = image.size
  core_width = float(max(adjusted_size, width))
  core_height = float(max(adjusted_size, height))

  scale_x = core_width / width
  scale_y = core_height / height
  scale = min(scale_x, scale_y)
  adjusted_width = width * scale
  adjusted_height = height * scale

  dx = (core_width - adjusted_width) / 2.0
  dy = (core_height - adjusted_height) / 2.0
  dx = round(dx) + border_size
  dy = round(dy) + border_size

  if scale != 1.0:
    width = round(width * scale)
    height = round(height * scale)
    image = image.resize((width, height), Resampling.BICUBIC)

  width = round(core_width) + 2 * border_size
  height = round(core_height) + 2 * border_size
  new_image = new("RGB", (width, height), (255, 255, 255))
  new_image.paste(image, (dx, dy))

  return new_image, dx, dy, scale

def _apply_fragment(rect: Rectangle, layout: Layout, dx: int, dy: int, scale: float):
  rect.lt = _apply_point(rect.lt, layout, dx, dy, scale)
  rect.lb = _apply_point(rect.lb, layout, dx, dy, scale)
  rect.rb = _apply_point(rect.rb, layout, dx, dy, scale)
  rect.rt = _apply_point(rect.rt, layout, dx, dy, scale)

def _apply_point(point: Point, layout: Layout, dx: int, dy: int, scale: float) -> Point:
  x, y = point
  x = (x - dx) / scale + layout.rect.lt[0]
  y = (y - dy) / scale + layout.rect.lt[1]
  return x, y

def _match_fragments(
    zone_rect: Rectangle,
    fragments1: Iterable[OCRFragment],
    fragments2: Iterable[OCRFragment],
  ) -> tuple[list[tuple[OCRFragment, OCRFragment]], list[OCRFragment]]:

  zone_polygon = Polygon(zone_rect)
  fragments2: list[OCRFragment] = list(fragments2)
  matched_fragments: list[tuple[OCRFragment, OCRFragment]] = []
  not_matched_fragments: list[OCRFragment] = []

  for fragment1 in fragments1:
    polygon1 = Polygon(fragment1.rect)
    polygon1 = zone_polygon.intersection(polygon1)
    if polygon1.is_empty:
      continue

    beast_j = -1
    beast_rate = 0.0

    for j, fragment2 in enumerate(fragments2):
      polygon2 = Polygon(fragment2.rect)
      rate = overlap_rate(polygon1, polygon2)
      if rate < _MIN_RATE:
        continue

      if rate > beast_rate:
        beast_j = j
        beast_rate = rate

    if beast_j != -1:
      matched_fragments.append((
        fragment1,
        fragments2[beast_j],
      ))
      del fragments2[beast_j]
    else:
      not_matched_fragments.append(fragment1)

  not_matched_fragments.extend(fragments2)
  return matched_fragments, not_matched_fragments


import os
import requests
from pathlib import Path


def download(url: str, file_path: Path):
  response = requests.get(url, stream=True, timeout=60)
  if response.status_code != 200:
    raise FileNotFoundError(f"Failed to download file from {url}: {response.status_code}")
  try:
    with open(file_path, "wb") as file:
      file.write(response.content)
  except Exception as e:
    if os.path.exists(file_path):
      os.remove(file_path)
    raise e



import numpy as np

from math import pi, ceil, sin, cos, sqrt
from PIL.Image import Image, Transform
from .types import Layout, ExtractedResult
from .rectangle import Rectangle
from .rotation import calculate_rotation_with_rect, normal_vertical_rotation


def clip(
    extracted_result: ExtractedResult,
    layout: Layout,
    wrapped_width: float = 0.0,
    wrapped_height: float = 0.0,
  ) -> Image:
  image: Image
  if extracted_result.adjusted_image is None:
    image = extracted_result.extracted_image
  else:
    image = extracted_result.adjusted_image

  return clip_from_image(
    image, layout.rect,
    wrapped_width, wrapped_height,
  )

def clip_from_image(
    image: Image,
    rect: Rectangle,
    wrapped_width: float = 0.0,
    wrapped_height: float = 0.0,
  ) -> Image:
  horizontal_rotation, vertical_rotation = calculate_rotation_with_rect(rect)
  image = image.copy()
  matrix_move = np.array(_get_move_matrix(rect.lt[0], rect.lt[1])).reshape(3, 3)
  matrix_rotate = np.array(_get_rotate_matrix(-horizontal_rotation)).reshape(3, 3)
  matrix = np.dot(matrix_move, matrix_rotate)

  y_axis_rotation = normal_vertical_rotation(vertical_rotation - horizontal_rotation)

  if abs(y_axis_rotation - 0.25 * pi) > 0.0:
    x = cos(y_axis_rotation)
    y = sin(y_axis_rotation)
    matrix_shear = np.array(_get_shear_matrix(x, y)).reshape(3, 3)
    matrix = np.dot(matrix, matrix_shear)

  width, height, max_width, max_height = _size_and_wrapper(rect)
  max_width += wrapped_width
  max_height += wrapped_height

  if max_width != width or max_height != height:
    dx = (max_width - width) / 2.0
    dy = (max_height - height) / 2.0
    matrix_move = np.array(_get_move_matrix(-dx, -dy)).reshape(3, 3)
    matrix = np.dot(matrix, matrix_move)

  return image.transform(
    size=(ceil(max_width), ceil(max_height)),
    method=Transform.AFFINE,
    data=_to_pillow_matrix(matrix),
  )

def _size_and_wrapper(rect: Rectangle):
  widths: list[float] = []
  heights: list[float] = []

  for i, (p1, p2) in enumerate(rect.segments):
    dx = p2[0] - p1[0]
    dy = p2[1] - p1[1]
    distance = sqrt(dx*dx + dy*dy)
    if i % 2 == 0:
      heights.append(distance)
    else:
      widths.append(distance)

  if len(widths) == 0 and len(heights) == 0:
    return 0.0, 0.0, 0.0, 0.0

  width: float = sum(widths) / len(widths)
  height: float = sum(heights) / len(heights)
  max_width: float = width
  max_height: float = height

  for width in widths:
    if width > max_width:
      max_width = width

  for height in heights:
    if height > max_height:
      max_height = height

  return width, height, max_width, max_height

def _to_pillow_matrix(matrix: np.array):
  return (
    matrix[0][0], matrix[0][1], matrix[0][2],
    matrix[1][0], matrix[1][1], matrix[1][2],
  )

def _get_move_matrix(dx: float, dy: float):
  return (
    1.0, 0.0, dx,
    0.0, 1.0, dy,
    0.0, 0.0, 1.0,
  )

def _get_rotate_matrix(rotation: float):
  return (
    cos(rotation),  sin(rotation),  0.0,
    -sin(rotation), cos(rotation),  0.0,
    0.0,            0.0,            1.0
  )

def _get_shear_matrix(x0: float, y0: float):
  return (
    1.0, 0.0, 0.0,
    x0,  y0,  0.0,
    0.0, 0.0, 1.0,
  )


import cv2
import numpy as np
import math
from PIL import Image


from .rec_postprocess import CTCLabelDecode
from .predict_base import PredictBase


class TextRecognizer(PredictBase):
  def __init__(self, args):
    self.rec_image_shape = args.rec_image_shape
    self.rec_batch_num = args.rec_batch_num
    self.rec_algorithm = args.rec_algorithm
    self.postprocess_op = CTCLabelDecode(
      character_dict_path=args.rec_char_dict_path,
      use_space_char=args.use_space_char,
    )

    # 初始化模型
    self.rec_onnx_session = self.get_onnx_session(args.rec_model_dir, args.use_gpu)
    self.rec_input_name = self.get_input_name(self.rec_onnx_session)
    self.rec_output_name = self.get_output_name(self.rec_onnx_session)

  def resize_norm_img(self, img, max_wh_ratio):
    imgC, imgH, imgW = self.rec_image_shape
    if self.rec_algorithm == "NRTR" or self.rec_algorithm == "ViTSTR":
      img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
      # return padding_im
      image_pil = Image.fromarray(np.uint8(img))
      if self.rec_algorithm == "ViTSTR":
        img = image_pil.resize([imgW, imgH], Image.BICUBIC)
      else:
        img = image_pil.resize([imgW, imgH], Image.ANTIALIAS)
      img = np.array(img)
      norm_img = np.expand_dims(img, -1)
      norm_img = norm_img.transpose((2, 0, 1))
      if self.rec_algorithm == "ViTSTR":
        norm_img = norm_img.astype(np.float32) / 255.0
      else:
        norm_img = norm_img.astype(np.float32) / 128.0 - 1.0
      return norm_img
    elif self.rec_algorithm == "RFL":
      img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
      resized_image = cv2.resize(img, (imgW, imgH), interpolation=cv2.INTER_CUBIC)
      resized_image = resized_image.astype("float32")
      resized_image = resized_image / 255
      resized_image = resized_image[np.newaxis, :]
      resized_image -= 0.5
      resized_image /= 0.5
      return resized_image

    assert imgC == img.shape[2]
    imgW = int((imgH * max_wh_ratio))

    # w = self.rec_onnx_session.get_inputs()[0].shape[3:][0]
    # w = self.rec_onnx_session.get_inputs()[0].shape[3:][0]
    # print(w)
    # if w is not None and w > 0:
    #   imgW = w

    h, w = img.shape[:2]
    ratio = w / float(h)
    if math.ceil(imgH * ratio) > imgW:
      resized_w = imgW
    else:
      resized_w = int(math.ceil(imgH * ratio))
    if self.rec_algorithm == "RARE":
      if resized_w > self.rec_image_shape[2]:
        resized_w = self.rec_image_shape[2]
      imgW = self.rec_image_shape[2]
    resized_image = cv2.resize(img, (resized_w, imgH))
    resized_image = resized_image.astype("float32")
    resized_image = resized_image.transpose((2, 0, 1)) / 255
    resized_image -= 0.5
    resized_image /= 0.5
    padding_im = np.zeros((imgC, imgH, imgW), dtype=np.float32)
    padding_im[:, :, 0:resized_w] = resized_image
    return padding_im

  def resize_norm_img_vl(self, img, image_shape):
    imgC, imgH, imgW = image_shape
    img = img[:, :, ::-1]  # bgr2rgb
    resized_image = cv2.resize(img, (imgW, imgH), interpolation=cv2.INTER_LINEAR)
    resized_image = resized_image.astype("float32")
    resized_image = resized_image.transpose((2, 0, 1)) / 255
    return resized_image

  def resize_norm_img_srn(self, img, image_shape):
    imgC, imgH, imgW = image_shape

    img_black = np.zeros((imgH, imgW))
    im_hei = img.shape[0]
    im_wid = img.shape[1]

    if im_wid <= im_hei * 1:
      img_new = cv2.resize(img, (imgH * 1, imgH))
    elif im_wid <= im_hei * 2:
      img_new = cv2.resize(img, (imgH * 2, imgH))
    elif im_wid <= im_hei * 3:
      img_new = cv2.resize(img, (imgH * 3, imgH))
    else:
      img_new = cv2.resize(img, (imgW, imgH))

    img_np = np.asarray(img_new)
    img_np = cv2.cvtColor(img_np, cv2.COLOR_BGR2GRAY)
    img_black[:, 0 : img_np.shape[1]] = img_np
    img_black = img_black[:, :, np.newaxis]

    row, col, c = img_black.shape
    c = 1

    return np.reshape(img_black, (c, row, col)).astype(np.float32)

  def srn_other_inputs(self, image_shape, num_heads, max_text_length):
    imgC, imgH, imgW = image_shape
    feature_dim = int((imgH / 8) * (imgW / 8))

    encoder_word_pos = (
      np.array(range(0, feature_dim)).reshape((feature_dim, 1)).astype("int64")
    )
    gsrm_word_pos = (
      np.array(range(0, max_text_length))
      .reshape((max_text_length, 1))
      .astype("int64")
    )

    gsrm_attn_bias_data = np.ones((1, max_text_length, max_text_length))
    gsrm_slf_attn_bias1 = np.triu(gsrm_attn_bias_data, 1).reshape(
      [-1, 1, max_text_length, max_text_length]
    )
    gsrm_slf_attn_bias1 = np.tile(gsrm_slf_attn_bias1, [1, num_heads, 1, 1]).astype(
      "float32"
    ) * [-1e9]

    gsrm_slf_attn_bias2 = np.tril(gsrm_attn_bias_data, -1).reshape(
      [-1, 1, max_text_length, max_text_length]
    )
    gsrm_slf_attn_bias2 = np.tile(gsrm_slf_attn_bias2, [1, num_heads, 1, 1]).astype(
      "float32"
    ) * [-1e9]

    encoder_word_pos = encoder_word_pos[np.newaxis, :]
    gsrm_word_pos = gsrm_word_pos[np.newaxis, :]

    return [
      encoder_word_pos,
      gsrm_word_pos,
      gsrm_slf_attn_bias1,
      gsrm_slf_attn_bias2,
    ]

  def process_image_srn(self, img, image_shape, num_heads, max_text_length):
    norm_img = self.resize_norm_img_srn(img, image_shape)
    norm_img = norm_img[np.newaxis, :]

    [encoder_word_pos, gsrm_word_pos, gsrm_slf_attn_bias1, gsrm_slf_attn_bias2] = (
      self.srn_other_inputs(image_shape, num_heads, max_text_length)
    )

    gsrm_slf_attn_bias1 = gsrm_slf_attn_bias1.astype(np.float32)
    gsrm_slf_attn_bias2 = gsrm_slf_attn_bias2.astype(np.float32)
    encoder_word_pos = encoder_word_pos.astype(np.int64)
    gsrm_word_pos = gsrm_word_pos.astype(np.int64)

    return (
      norm_img,
      encoder_word_pos,
      gsrm_word_pos,
      gsrm_slf_attn_bias1,
      gsrm_slf_attn_bias2,
    )

  def resize_norm_img_sar(self, img, image_shape, width_downsample_ratio=0.25):
    imgC, imgH, imgW_min, imgW_max = image_shape
    h = img.shape[0]
    w = img.shape[1]
    valid_ratio = 1.0
    # make sure new_width is an integral multiple of width_divisor.
    width_divisor = int(1 / width_downsample_ratio)
    # resize
    ratio = w / float(h)
    resize_w = math.ceil(imgH * ratio)
    if resize_w % width_divisor != 0:
      resize_w = round(resize_w / width_divisor) * width_divisor
    if imgW_min is not None:
      resize_w = max(imgW_min, resize_w)
    if imgW_max is not None:
      valid_ratio = min(1.0, 1.0 * resize_w / imgW_max)
      resize_w = min(imgW_max, resize_w)
    resized_image = cv2.resize(img, (resize_w, imgH))
    resized_image = resized_image.astype("float32")
    # norm
    if image_shape[0] == 1:
      resized_image = resized_image / 255
      resized_image = resized_image[np.newaxis, :]
    else:
      resized_image = resized_image.transpose((2, 0, 1)) / 255
    resized_image -= 0.5
    resized_image /= 0.5
    resize_shape = resized_image.shape
    padding_im = -1.0 * np.ones((imgC, imgH, imgW_max), dtype=np.float32)
    padding_im[:, :, 0:resize_w] = resized_image
    pad_shape = padding_im.shape

    return padding_im, resize_shape, pad_shape, valid_ratio

  def resize_norm_img_spin(self, img):
    img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    # return padding_im
    img = cv2.resize(img, tuple([100, 32]), cv2.INTER_CUBIC)
    img = np.array(img, np.float32)
    img = np.expand_dims(img, -1)
    img = img.transpose((2, 0, 1))
    mean = [127.5]
    std = [127.5]
    mean = np.array(mean, dtype=np.float32)
    std = np.array(std, dtype=np.float32)
    mean = np.float32(mean.reshape(1, -1))
    stdinv = 1 / np.float32(std.reshape(1, -1))
    img -= mean
    img *= stdinv
    return img

  def resize_norm_img_svtr(self, img, image_shape):
    imgC, imgH, imgW = image_shape
    resized_image = cv2.resize(img, (imgW, imgH), interpolation=cv2.INTER_LINEAR)
    resized_image = resized_image.astype("float32")
    resized_image = resized_image.transpose((2, 0, 1)) / 255
    resized_image -= 0.5
    resized_image /= 0.5
    return resized_image

  def resize_norm_img_abinet(self, img, image_shape):
    imgC, imgH, imgW = image_shape

    resized_image = cv2.resize(img, (imgW, imgH), interpolation=cv2.INTER_LINEAR)
    resized_image = resized_image.astype("float32")
    resized_image = resized_image / 255.0

    mean = np.array([0.485, 0.456, 0.406])
    std = np.array([0.229, 0.224, 0.225])
    resized_image = (resized_image - mean[None, None, ...]) / std[None, None, ...]
    resized_image = resized_image.transpose((2, 0, 1))
    resized_image = resized_image.astype("float32")

    return resized_image

  def norm_img_can(self, img, image_shape):
    img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)  # CAN only predict gray scale image

    if self.inverse:
      img = 255 - img

    if self.rec_image_shape[0] == 1:
      h, w = img.shape
      _, imgH, imgW = self.rec_image_shape
      if h < imgH or w < imgW:
        padding_h = max(imgH - h, 0)
        padding_w = max(imgW - w, 0)
        img_padded = np.pad(
          img,
          ((0, padding_h), (0, padding_w)),
          "constant",
          constant_values=(255),
        )
        img = img_padded

    img = np.expand_dims(img, 0) / 255.0  # h,w,c -> c,h,w
    img = img.astype("float32")

    return img

  def __call__(self, img_list):
    img_num = len(img_list)
    # Calculate the aspect ratio of all text bars
    width_list = []
    for img in img_list:
      width_list.append(img.shape[1] / float(img.shape[0]))
    # Sorting can speed up the recognition process
    indices = np.argsort(np.array(width_list))
    rec_res = [["", 0.0]] * img_num
    batch_num = self.rec_batch_num

    for beg_img_no in range(0, img_num, batch_num):
      end_img_no = min(img_num, beg_img_no + batch_num)
      norm_img_batch = []
      imgC, imgH, imgW = self.rec_image_shape[:3]
      max_wh_ratio = imgW / imgH
      # max_wh_ratio = 0
      for ino in range(beg_img_no, end_img_no):
        h, w = img_list[indices[ino]].shape[0:2]
        wh_ratio = w * 1.0 / h
        max_wh_ratio = max(max_wh_ratio, wh_ratio)
      for ino in range(beg_img_no, end_img_no):
        norm_img = self.resize_norm_img(img_list[indices[ino]], max_wh_ratio)
        norm_img = norm_img[np.newaxis, :]
        norm_img_batch.append(norm_img)

      norm_img_batch = np.concatenate(norm_img_batch)
      norm_img_batch = norm_img_batch.copy()

      # img = img[:, :, ::-1].transpose(2, 0, 1)
      # img = img[:, :, ::-1]
      # img = img.transpose(2, 0, 1)
      # img = img.astype(np.float32)
      # img = np.expand_dims(img, axis=0)
      # print(img.shape)
      input_feed = self.get_input_feed(self.rec_input_name, norm_img_batch)
      outputs = self.rec_onnx_session.run(
        self.rec_output_name, input_feed=input_feed
      )

      preds = outputs[0]

      rec_result = self.postprocess_op(preds)
      for rno in range(len(rec_result)):
        rec_res[indices[beg_img_no + rno]] = rec_result[rno]

    return rec_res



import torch

from torch import nn
from transformers import AutoModelForVision2Seq, AutoProcessor


class Pix2Struct(nn.Module):
    def __init__(
            self,
            model_path='U4R/StructTable-base',
            max_new_tokens=1024,
            max_time=30,
            cache_dir=None,
            local_files_only=None,
            **kwargs,
        ):
        super().__init__()
        self.model_path = model_path
        self.max_new_tokens = max_new_tokens
        self.max_generate_time = max_time
        self.cache_dir = cache_dir
        self.local_files_only = local_files_only

        # init model and image processor from ckpt path
        self.init_image_processor(model_path)
        self.init_model(model_path)

        self.special_str_list = ['\\midrule', '\\hline']
        self.supported_output_format = ['latex']

    def postprocess_latex_code(self, code):
        for special_str in self.special_str_list:
            code = code.replace(special_str, special_str + ' ')
        return code

    def init_model(self, model_path):
        self.model = AutoModelForVision2Seq.from_pretrained(
            pretrained_model_name_or_path=model_path,
            cache_dir=self.cache_dir,
            local_files_only=self.local_files_only,
        )
        self.model.eval()

    def init_image_processor(self, image_processor_path):
        self.data_processor = AutoProcessor.from_pretrained(
            pretrained_model_name_or_path=image_processor_path,
            cache_dir=self.cache_dir,
            local_files_only=self.local_files_only,
        )

    def forward(self, image, **kwargs):
        # process image to tokens
        image_tokens = self.data_processor.image_processor(
            images=image,
            return_tensors='pt',
        )

        device = next(self.parameters()).device
        for k, v in image_tokens.items():
            image_tokens[k] = v.to(device)

        # generate text from image tokens
        model_output = self.model.generate(
            flattened_patches=image_tokens['flattened_patches'],
            attention_mask=image_tokens['attention_mask'],
            max_new_tokens=self.max_new_tokens,
            max_time=self.max_generate_time,
            no_repeat_ngram_size=20,
        )

        latex_codes = self.data_processor.batch_decode(model_output, skip_special_tokens=True)
        # postprocess
        for i, code in enumerate(latex_codes):
            latex_codes[i] = self.postprocess_latex_code(code)

        return latex_codes



import os

from logging import Logger
from huggingface_hub import hf_hub_download, snapshot_download, try_to_load_from_cache
from .types import ModelsDownloader

class HuggingfaceModelsDownloader(ModelsDownloader):
  def __init__(
      self,
      logger: Logger,
      model_dir_path: str | None
    ):
    self._logger = logger
    self._model_dir_path: str | None = model_dir_path

  def onnx_ocr(self) -> str:
    repo_path = try_to_load_from_cache(
      repo_id="moskize/OnnxOCR",
      filename="README.md",
      cache_dir=self._model_dir_path
    )
    if isinstance(repo_path, str):
      return os.path.dirname(repo_path)
    else:
      self._logger.info("Downloading OCR model...")
      return snapshot_download(
        cache_dir=self._model_dir_path,
        repo_id="moskize/OnnxOCR",
      )

  def yolo(self) -> str:
    yolo_file_path = try_to_load_from_cache(
      repo_id="opendatalab/PDF-Extract-Kit-1.0",
      filename="models/Layout/YOLO/doclayout_yolo_ft.pt",
      cache_dir=self._model_dir_path
    )
    if isinstance(yolo_file_path, str):
      return yolo_file_path
    else:
      self._logger.info("Downloading YOLO model...")
      return hf_hub_download(
        cache_dir=self._model_dir_path,
        repo_id="opendatalab/PDF-Extract-Kit-1.0",
        filename="models/Layout/YOLO/doclayout_yolo_ft.pt",
      )

  def layoutreader(self) -> str:
    repo_path = try_to_load_from_cache(
      repo_id="hantian/layoutreader",
      filename="model.safetensors",
      cache_dir=self._model_dir_path
    )
    if isinstance(repo_path, str):
      return os.path.dirname(repo_path)
    else:
      self._logger.info("Downloading LayoutReader model...")
      return snapshot_download(
        cache_dir=self._model_dir_path,
        repo_id="hantian/layoutreader",
      )

  def struct_eqtable(self) -> str:
    repo_path = try_to_load_from_cache(
      repo_id="U4R/StructTable-InternVL2-1B",
      filename="model.safetensors",
      cache_dir=self._model_dir_path
    )
    if isinstance(repo_path, str):
      return os.path.dirname(repo_path)
    else:
      self._logger.info("Downloading StructEqTable model...")
      return snapshot_download(
        cache_dir=self._model_dir_path,
        repo_id="U4R/StructTable-InternVL2-1B",
      )

  def latex(self):
    repo_path = try_to_load_from_cache(
      repo_id="lukbl/LaTeX-OCR",
      filename="checkpoints/weights.pth",
      repo_type="space",
      cache_dir=self._model_dir_path
    )
    if isinstance(repo_path, str):
      return os.path.dirname(os.path.dirname(repo_path))
    else:
      self._logger.info("Downloading LaTeX model...")
      return snapshot_download(
        cache_dir=self._model_dir_path,
        repo_type="space",
        repo_id="lukbl/LaTeX-OCR",
      )



from setuptools import setup, find_packages

if "doc_page_extractor.struct_eqtable" not in find_packages():
  raise RuntimeError("struct_eqtable not found. Please download struct_eqtable first.")

setup(
  name="doc-page-extractor-test",
  version="0.1.4",
  author="Tao Zeyu",
  author_email="i@taozeyu.com",
  url="https://github.com/Moskize91/doc-page-extractor",
  description="doc page extractor can identify text and format in images and return structured data.",
  packages=find_packages(),
  long_description=open("./README.md", encoding="utf8").read(),
  long_description_content_type="text/markdown",
  install_requires=[
    "opencv-python>=4.11.0,<5.0",
    "pillow>=10.3,<11.0",
    "pyclipper>=1.2.0,<2.0",
    "numpy>=1.24.0,<2.0",
    "shapely>=2.0.0,<3.0",
    "transformers>=4.42.4,<=4.47",
    "doclayout_yolo>=0.0.3",
    "pix2tex>=0.1.4,<=0.2.0",
    "accelerate>=1.6.0,<2.0",
    "huggingface_hub>=0.30.2",
  ],
)


import numpy as np
from .imaug import transform, create_operators
from .db_postprocess import DBPostProcess
from .predict_base import PredictBase


class TextDetector(PredictBase):
  def __init__(self, args):
    self.args = args
    self.det_algorithm = args.det_algorithm
    pre_process_list = [
      {
        "DetResizeForTest": {
          "limit_side_len": args.det_limit_side_len,
          "limit_type": args.det_limit_type,
        }
      },
      {
        "NormalizeImage": {
          "std": [0.229, 0.224, 0.225],
          "mean": [0.485, 0.456, 0.406],
          "scale": "1./255.",
          "order": "hwc",
        }
      },
      {"ToCHWImage": None},
      {"KeepKeys": {"keep_keys": ["image", "shape"]}},
    ]
    postprocess_params = {}
    postprocess_params["name"] = "DBPostProcess"
    postprocess_params["thresh"] = args.det_db_thresh
    postprocess_params["box_thresh"] = args.det_db_box_thresh
    postprocess_params["max_candidates"] = 1000
    postprocess_params["unclip_ratio"] = args.det_db_unclip_ratio
    postprocess_params["use_dilation"] = args.use_dilation
    postprocess_params["score_mode"] = args.det_db_score_mode
    postprocess_params["box_type"] = args.det_box_type

    # 实例化预处理操作类
    self.preprocess_op = create_operators(pre_process_list)
    # self.postprocess_op = build_post_process(postprocess_params)
    # 实例化后处理操作类
    self.postprocess_op = DBPostProcess(**postprocess_params)

    # 初始化模型
    self.det_onnx_session = self.get_onnx_session(args.det_model_dir, args.use_gpu)
    self.det_input_name = self.get_input_name(self.det_onnx_session)
    self.det_output_name = self.get_output_name(self.det_onnx_session)

  def order_points_clockwise(self, pts):
    rect = np.zeros((4, 2), dtype="float32")
    s = pts.sum(axis=1)
    rect[0] = pts[np.argmin(s)]
    rect[2] = pts[np.argmax(s)]
    tmp = np.delete(pts, (np.argmin(s), np.argmax(s)), axis=0)
    diff = np.diff(np.array(tmp), axis=1)
    rect[1] = tmp[np.argmin(diff)]
    rect[3] = tmp[np.argmax(diff)]
    return rect

  def clip_det_res(self, points, img_height, img_width):
    for pno in range(points.shape[0]):
      points[pno, 0] = int(min(max(points[pno, 0], 0), img_width - 1))
      points[pno, 1] = int(min(max(points[pno, 1], 0), img_height - 1))
    return points

  def filter_tag_det_res(self, dt_boxes, image_shape):
    img_height, img_width = image_shape[0:2]
    dt_boxes_new = []
    for box in dt_boxes:
      if type(box) is list:
        box = np.array(box)
      box = self.order_points_clockwise(box)
      box = self.clip_det_res(box, img_height, img_width)
      rect_width = int(np.linalg.norm(box[0] - box[1]))
      rect_height = int(np.linalg.norm(box[0] - box[3]))
      if rect_width <= 3 or rect_height <= 3:
        continue
      dt_boxes_new.append(box)
    dt_boxes = np.array(dt_boxes_new)
    return dt_boxes

  def filter_tag_det_res_only_clip(self, dt_boxes, image_shape):
    img_height, img_width = image_shape[0:2]
    dt_boxes_new = []
    for box in dt_boxes:
      if type(box) is list:
        box = np.array(box)
      box = self.clip_det_res(box, img_height, img_width)
      dt_boxes_new.append(box)
    dt_boxes = np.array(dt_boxes_new)
    return dt_boxes

  def __call__(self, img):
    ori_im = img.copy()
    data = {"image": img}

    data = transform(data, self.preprocess_op)
    img, shape_list = data
    if img is None:
      return None, 0
    img = np.expand_dims(img, axis=0)
    shape_list = np.expand_dims(shape_list, axis=0)
    img = img.copy()

    input_feed = self.get_input_feed(self.det_input_name, img)
    outputs = self.det_onnx_session.run(self.det_output_name, input_feed=input_feed)

    preds = {}
    preds["maps"] = outputs[0]

    post_result = self.postprocess_op(preds, shape_list)
    dt_boxes = post_result[0]["points"]

    if self.args.det_box_type == "poly":
      dt_boxes = self.filter_tag_det_res_only_clip(dt_boxes, ori_im.shape)
    else:
      dt_boxes = self.filter_tag_det_res(dt_boxes, ori_im.shape)

    return dt_boxes



import numpy as np

from dataclasses import dataclass
from PIL.Image import Image
from math import pi
from .types import OCRFragment, Layout
from .rotation import calculate_rotation, RotationAdjuster
from .rectangle import Rectangle


_TINY_ROTATION = 0.005 # below this angle, we consider the text is horizontal


@dataclass
class _RotationContext:
  to_origin: RotationAdjuster
  to_new: RotationAdjuster
  fragment_origin_rectangles: list[Rectangle]

class RawOptimizer:
  def __init__(
      self,
      raw: Image,
      adjust_points: bool,
    ):
    self._raw: Image = raw
    self._image: Image = raw
    self._adjust_points: bool = adjust_points
    self._fragments: list[OCRFragment]
    self._rotation: float = 0.0
    self._rotation_context: _RotationContext | None  = None

  @property
  def image(self) -> Image:
    return self._image

  @property
  def adjusted_image(self) -> Image | None:
    if self._adjust_points and self._image != self._raw:
      return self._image

  @property
  def rotation(self) -> float:
    return self._rotation

  @property
  def image_np(self) -> np.ndarray:
    return np.array(self._raw)

  def receive_raw_fragments(self, fragments: list[OCRFragment]):
    self._fragments = fragments
    self._rotation = calculate_rotation(fragments)

    if abs(self._rotation) < _TINY_ROTATION:
      return

    origin_size = self._raw.size
    self._image = self._raw.rotate(
      angle=self._rotation * 180 / pi,
      fillcolor=(255, 255, 255),
      expand=True,
    )
    self._rotation_context = _RotationContext(
      fragment_origin_rectangles=[f.rect for f in fragments],
      to_origin=RotationAdjuster(
        origin_size=origin_size,
        new_size=self._image.size,
        rotation=self._rotation,
        to_origin_coordinate=True,
      ),
      to_new=RotationAdjuster(
        origin_size=origin_size,
        new_size=self._image.size,
        rotation=self._rotation,
        to_origin_coordinate=False,
      ),
    )
    adjuster = self._rotation_context.to_new

    for fragment in fragments:
      rect = fragment.rect
      fragment.rect = Rectangle(
        lt=adjuster.adjust(rect.lt),
        rt=adjuster.adjust(rect.rt),
        lb=adjuster.adjust(rect.lb),
        rb=adjuster.adjust(rect.rb),
      )

  def receive_raw_layouts(self, layouts: list[Layout]):
    if self._adjust_points or self._rotation_context is None:
      return

    for fragment, origin_rect in zip(self._fragments, self._rotation_context.fragment_origin_rectangles):
      fragment.rect = origin_rect

    adjuster = self._rotation_context.to_origin

    for layout in layouts:
      layout.rect = Rectangle(
        lt=adjuster.adjust(layout.rect.lt),
        rt=adjuster.adjust(layout.rect.rt),
        lb=adjuster.adjust(layout.rect.lb),
        rb=adjuster.adjust(layout.rect.rb),
      )



import os
import torch

from munch import Munch
from pix2tex.cli import LatexOCR
from PIL.Image import Image
from typing import Literal
from .utils import expand_image
from .types import GetModelDir

class LaTeX:
  def __init__(self, device: Literal["cpu", "cuda"],get_model_dir: GetModelDir):
    self._model_path: str = get_model_dir()
    self._model: LatexOCR | None = None
    self._device: Literal["cpu", "cuda"] = device

  def extract(self, image: Image) -> str | None:
    image = expand_image(image, 0.1) # 添加边缘提高识别准确率
    model = self._get_model()
    with torch.no_grad():
      return model(image)

  def _get_model(self) -> LatexOCR:
    if self._model is None:
      self._model = LatexOCR(Munch({
        "config": os.path.join("settings", "config.yaml"),
        "checkpoint": os.path.join(self._model_path, "checkpoints", "weights.pth"),
        "no_cuda": self._device == "cpu",
        "no_resize": False,
      }))
    return self._model



import cv2
import copy
import numpy as np
import math

from .cls_postprocess import ClsPostProcess
from .predict_base import PredictBase


class TextClassifier(PredictBase):
  def __init__(self, args):
    self.cls_image_shape = args.cls_image_shape
    self.cls_batch_num = args.cls_batch_num
    self.cls_thresh = args.cls_thresh
    self.postprocess_op = ClsPostProcess(label_list=args.label_list)

    # 初始化模型
    self.cls_onnx_session = self.get_onnx_session(args.cls_model_dir, args.use_gpu)
    self.cls_input_name = self.get_input_name(self.cls_onnx_session)
    self.cls_output_name = self.get_output_name(self.cls_onnx_session)

  def resize_norm_img(self, img):
    imgC, imgH, imgW = self.cls_image_shape
    h = img.shape[0]
    w = img.shape[1]
    ratio = w / float(h)
    if math.ceil(imgH * ratio) > imgW:
      resized_w = imgW
    else:
      resized_w = int(math.ceil(imgH * ratio))
    resized_image = cv2.resize(img, (resized_w, imgH))
    resized_image = resized_image.astype("float32")
    if self.cls_image_shape[0] == 1:
      resized_image = resized_image / 255
      resized_image = resized_image[np.newaxis, :]
    else:
      resized_image = resized_image.transpose((2, 0, 1)) / 255
    resized_image -= 0.5
    resized_image /= 0.5
    padding_im = np.zeros((imgC, imgH, imgW), dtype=np.float32)
    padding_im[:, :, 0:resized_w] = resized_image
    return padding_im

  def __call__(self, img_list):
    img_list = copy.deepcopy(img_list)
    img_num = len(img_list)
    # Calculate the aspect ratio of all text bars
    width_list = []
    for img in img_list:
      width_list.append(img.shape[1] / float(img.shape[0]))
    # Sorting can speed up the cls process
    indices = np.argsort(np.array(width_list))

    cls_res = [["", 0.0]] * img_num
    batch_num = self.cls_batch_num

    for beg_img_no in range(0, img_num, batch_num):

      end_img_no = min(img_num, beg_img_no + batch_num)
      norm_img_batch = []
      max_wh_ratio = 0

      for ino in range(beg_img_no, end_img_no):
        h, w = img_list[indices[ino]].shape[0:2]
        wh_ratio = w * 1.0 / h
        max_wh_ratio = max(max_wh_ratio, wh_ratio)
      for ino in range(beg_img_no, end_img_no):
        norm_img = self.resize_norm_img(img_list[indices[ino]])
        norm_img = norm_img[np.newaxis, :]
        norm_img_batch.append(norm_img)
      norm_img_batch = np.concatenate(norm_img_batch)
      norm_img_batch = norm_img_batch.copy()

      input_feed = self.get_input_feed(self.cls_input_name, norm_img_batch)
      outputs = self.cls_onnx_session.run(
        self.cls_output_name, input_feed=input_feed
      )

      prob_out = outputs[0]

      cls_result = self.postprocess_op(prob_out)
      for rno in range(len(cls_result)):
        label, score = cls_result[rno]
        cls_res[indices[beg_img_no + rno]] = [label, score]
        if "180" in label and score > self.cls_thresh:
          img_list[indices[beg_img_no + rno]] = cv2.rotate(
            img_list[indices[beg_img_no + rno]], 1
          )
    return img_list, cls_res
